(* Copyright 2006, 2008 Andrew W. Appel, Robert Dockins and Xavier Leroy *)
 * This file is part of the List-machine Benchmark.
 * The List-machine Benchmark is licensed under GPLv3; see LICENSE.
 * )

Require Export LogicRules.
Require Export ModalRules.
Require Export OtherRules.

Open Local Scope ty_scope.

Lemma ignores_incl : forall l l' t,
  incl l' l ->
  ignores l t ->
  ignores l' t.

  unfold ignores; intros.
  cut_with (ignoring l t); auto.
  apply_d (ignore_incl l); auto.
  assumed.
 Qed.  

Lemma top_ignores : forall l,
  ignores l t_top.

   unfold ignores; intros.
   apply derives_axiom.
   apply modal_N.
   apply top_intro.
 Qed.

Lemma bot_ignores : forall l,
  ignores l t_bot.

   unfold ignores; intros.
   elim_at 0 bot_elim.
  Qed.

Lemma const_ignores : forall l P,
  ignores l (t_const P).

  unfold ignores; intros.
  rewrite ignoring_spec; unfold derives, t_ignoring, t_const; intros.
  pattern w; refine (H _ _); simpl; auto.
 Qed.
  
Lemma intersection_ignores : forall l t1 t2,
  ignores l t1 -> 
  ignores l t2 ->
  ignores l (t1 && t2).

  unfold ignores; intros.
  simpl_derives.
  weaken_with H.
  weaken_with H0.
 Qed.

Lemma union_ignores : forall l t1 t2,
  ignores l t1 -> 
  ignores l t2 ->
  ignores l (t1 || t2).

  unfold ignores; intros.
  elim_at 0 union_elim.
  cut_with (ignoring l t1); [ | weaken_with H ].
  coalesce_modal; apply union_intro1; assumed.
  cut_with (ignoring l t2); [ | weaken_with H0 ].
  coalesce_modal; apply union_intro2; assumed.
 Qed.

Lemma implies_ignores : forall l t1 t2,
  ignores l t1 ->
  ignores l t2 ->
  ignores l (t1 --> t2).

  unfold ignores; intros l t1 t2.
  rewrite ignoring_spec.
  unfold derives; intros.
  apply ignoring'_ignoring.
  unfold t_ignoring', t_with_subst; intros.
  red; intros.
  unfold t_ignoring in *.
  apply H0 with w; auto.
  simpl; intros; intuition; subst t'.
  cut ((t1 --> t2) w); [ | refine (H1 _ _); simpl; auto ].
  unfold t_implies; intro H4; apply H4.
  apply H with (world_update z w).
  simpl; intros; intuition; subst t'.
  auto.
  intros; rewrite (proj_subst_notin x z l w); auto.
  intros; rewrite (proj_subst_notin x z l w); auto.
 Qed.
  
Lemma forall_ignores : forall l A (f:A->ty),
  (forall a:A, ignores l (f a)) ->
  ignores l (t_all f).

  unfold ignores; intros.
  rewrite ignoring_spec.
  unfold derives; intros.
  cut (t_all f w); [ | refine (H0 _ _); simpl; auto ].
  unfold t_all; intros.
  apply ignoring'_ignoring.
  unfold t_ignoring', t_with_subst; intros.
  unfold derives in H.
  rewrite ignoring_spec in H; unfold t_ignoring in H.  
  apply H with w.
  simpl; intros; intuition; subst t'.
  apply H1.
  intros.
  rewrite (proj_subst_notin x0 z l w); auto.
 Qed.

Lemma exist_ignores : forall l A (f:A->ty),
  (forall a:A, ignores l (f a)) ->
  ignores l (t_ex f).

  unfold ignores; intros.
  simpl_derives.
  apply_d modal_distrib_exist.
  apply exist_intro with x.
  apply H.
 Qed.

Lemma later_ignores : forall l t,
  ~In Sel_N l ->
  ignores l t ->
  ignores l (|>t).

  unfold ignores; intros.
  rewrite ignore_later_commute; auto.
  apply derives_modal; auto.
 Qed.

Lemma nec_ignores : forall l t,
  ~In Sel_N l ->
  ignores l t ->
  ignores l ([]t).

  unfold ignores; intros.
  rewrite ignore_nec_commute; auto.
  apply derives_modal; auto.
 Qed.

Lemma ignoring_ignores : forall l l' t,
  ignores l t ->
  ignores (l++l') (ignoring l' t).

  unfold ignores; intros.  
  rewrite ignore_app.
  rewrite <- ignore_ignore.
  rewrite ignore_commute.
  coalesce_modal; auto.
 Qed.  

Lemma ignoring_ignores_immediate : forall l l' t,
  incl l l' ->
  ignores l (ignoring l' t).

  unfold ignores; intros.
  rewrite <- ignore_app.
  apply ignore_incl.
  red; intros.
  elim (in_app_or _ _ _ H0).
  apply H.
  auto.
 Qed.


Lemma with_subst_ignores : forall l z t,
  ignores l t ->
  ignores l (with_subst z t).

  unfold ignores; intros.
  apply_d with_subst_ignoring.
  coalesce_modal; auto.
 Qed.

Lemma projecting_ignores : forall l x a,
  ~In x l ->
  ignores l (projecting x a).

  unfold ignores; intros.
  rewrite ignoring_spec; unfold derives; intros.
  cut (projecting x a w); [ | refine (H0 _ _); simpl; auto ].
  unfold projecting, t_ignoring; intros.
  subst a; rewrite H2; auto.
 Qed.

Lemma comp_prop_ignores : forall l x f,
  ~In x l ->
  (forall a, ignores l (f a)) ->
  ignores l (t_comp_prop x f).

  unfold ignores; intros.
  rewrite ignoring_spec; unfold derives; intros.
  apply ignoring'_ignoring.
  unfold t_ignoring', t_with_subst; intros.
  unfold t_comp_prop; simpl.
  rewrite ignoring_spec in H0.
  unfold derives, t_ignoring in H0.
  apply H0 with w.
  simpl; intros; intuition; subst t'.
  cut (t_comp_prop x f w); [ | refine (H1 _ _); simpl; auto ].
  unfold t_comp_prop.
  rewrite (proj_subst_notin x z l w); auto.
  intros.
  rewrite (proj_subst_notin x0 z l w); auto.  
 Qed.


Lemma ignores_app : forall l1 l2 t,
  ignores l1 t ->
  ignores l2 t ->
  ignores (l1++l2) t.

  intros.
  unfold ignores.
  rewrite ignore_app.
  cut_with (ignoring l1 t).
  coalesce_modal.
  apply H0.
  apply H.
 Qed.

Lemma with_subst_ignores_upd : forall z t,
  ignores (upd_set z) (with_subst z t).

  unfold ignores.
  intros.
  unfold derives; intros.
  cut (with_subst z t w); [ | refine (H _ _); simpl; auto ].
  clear H.
  rewrite with_subst_spec.  
  rewrite ignoring_spec.
  intros.
  apply ignoring'_ignoring.
  red; intros.
  repeat (red; simpl).  
  rewrite world_update_drop.
  apply H.
  auto.
 Qed.

Lemma slot_ignores : forall l x t, 
   ~In Sel_S l ->
   ignores l t ->
   ignores l (t_slot x t).
   
   intros; red.
   unfold t_slot.
   apply exist_ignores; intros.
   apply intersection_ignores.
   apply comp_prop_ignores; auto; intros.
   intros; apply const_ignores.
   apply with_subst_ignores; auto.
  Qed.   


Lemma slot_ignores_V : forall x t,
   ignores (Sel_V::nil) (t_slot x t).

   intros; red.
   unfold t_slot.
   apply exist_ignores; intros.
   apply intersection_ignores.
   apply comp_prop_ignores; auto; intros.
   simpl; intuition; discriminate.
   apply const_ignores.
   refine (with_subst_ignores_upd _ _).
 Qed.


Lemma slot_nec : forall x t,
   t |-- |> t ->
   t_slot x t |-- |>(t_slot x t).

   intros; unfold t_slot.
   simpl_derives.
   apply_d modal_distrib_exist.
   apply exist_intro with x0.
   simpl_derives.
   apply_d (ignore_later (Sel_N::nil)).
   clear_at 1.
   apply comp_prop_ignores.
   simpl; intuition; discriminate.
   intros; apply const_ignores.
   simpl; intuition.
   clear_at 0.
   rewrite <- with_subst_later_commute.
   coalesce_modal.
   auto.
   simpl; intuition; discriminate.
  Qed.


 Lemma slot_sub : forall x t1 t2,
   t1 >=> t2 |-- t_slot x t1 >=> t_slot x t2.

   intros; unfold t_sub, t_slot.
   cut_with ([]!! (t1 --> t2)).
   clear_at 1.
   do 2 coalesce_modal.
   apply (with_projecting Sel_S); intro s.
   elim_at 1 (comp_prop_elim _ s); [ assumed | ].
   elim_at 0 const_elim; intros.
   apply exist_intro with x0.
   simpl_derives.
   apply comp_prop_intro with s; [ assumed | ].
   apply const_intro; auto.
   cut_with (with_subst (existT _ Sel_V x0::nil) (!(t1 --> t2))).
   coalesce_modal.
   cut_with (t1 --> t2).
   elim_at 0 implies_elim; assumed.
   unfold t_every.
   clear_at 0.
   apply ignore_any.
   move_at 2.
   do 2 clear_at 1.
   rewrite discard_with_subst.
   assumed.
   simpl; unfold t_every.
   apply ignoring_ignores_immediate.
   simpl; intuition.
   unfold t_every; rewrite <- ignore_ignore.
   assumed.
  Qed.


Hint Resolve top_ignores bot_ignores const_ignores
  intersection_ignores union_ignores
  implies_ignores forall_ignores
  exist_ignores later_ignores nec_ignores
  ignoring_ignores with_subst_ignores
  projecting_ignores comp_prop_ignores
  : ignores.
