(* Copyright 2006, 2008 Andrew W. Appel, Robert Dockins and Xavier Leroy *)
 * This file is part of the List-machine Benchmark.
 * The List-machine Benchmark is licensed under GPLv3; see LICENSE.
 * )

Require Export TC_syntax.

Open Local Scope ty_scope.

Lemma interp_valuepred : forall p t,
  [[ t |p] = t_valuepred [[ t |p].

  intros p t; unfold t_valuepred.
  simpl; unfold t_compose.
  apply derives_equal.
  apply nec_split.
  destruct (interp_ignore_nec p (1+list_ty_depth t)).
  destruct (H0 t); auto.
  cut_with (|> [[ t |p]).
  coalesce_modal.
  destruct (interp_ignore_nec p (1+list_ty_depth t)).
  destruct (H0 t); auto.
  destruct (interp_ignore_nec p (1+list_ty_depth t)).
  destruct (H0 t); auto.
  apply_d nec_now.
  apply derives_modal.
  apply ignore_any.
 Qed.

Lemma later_interp_valuepred : forall p t,
  |>[[ t |p] = t_valuepred (|>[[ t |p]).

  intros; unfold t_valuepred.
  simpl; unfold t_compose.
  rewrite ignore_later_commute.  
2:  simpl; intuition; discriminate.
  rewrite nec_later_later.
  rewrite interp_valuepred.
  simpl.
  unfold t_compose.
  rewrite ignore_nec_commute.
2:  simpl; intuition; discriminate.
  rewrite <- ignore_ignore.
  auto.
 Qed.

Lemma lm_check_block_jump: forall p e c v pt,
   cm_get e v = Some (lty_cont c) ->
   |-- interp_cont p e 1 >=> interp_cont p c 1 ->
   check_block p (lm_pt p pt) (lm_ok p pt) (interp_cont p e 1) (instr_jump v).

   intros.
   apply check_block_weaken with
     (interp_cont p c 1 && (t_slot v (t_cont p (interp_cont p c 1)))).
   cut_with (interp_cont p e 1 >=> interp_cont p c 1).
   cut_with (interp_cont p e 1 >=> t_slot v (t_cont p (interp_cont p c 1))).
   clear_at 2.
   unfold t_sub; do 2 coalesce_modal.
   elim_at 1 implies_elim; assumed.
   elim_at 0 implies_elim; assumed.
   apply derives_axiom.
   apply (cm_get_interp p e v (lty_cont c)).
   auto.
   apply derives_axiom.
   apply H0.

   apply check_block_jump.
   apply lm_pt_every.
   apply lm_ok_nec_every.
   destruct (interp_ignore_nec p (1 + contmap_depth c)).
   destruct (H1 c 1%positive); auto.
   destruct (interp_ignore_nec p (1 + contmap_depth c)).
   destruct (H1 c 1%positive); auto.
 Qed.


Lemma lm_check_instr_cons : forall p pt e v v0 v1 t0 t1 t',
  cm_get e v0 = Some t0 ->
  cm_get e v1 = Some t1 ->
  |-- [[ t0 |p] >=> [[ t' |p] ->
  |-- [[ t1 |p] >=> [[ lty_list t' |p] ->
  check_instr p (lm_pt p pt) (lm_ok p pt) (interp_cont p e 1)
     (instr_cons v0 v1 v) (interp_cont p (cm_set e v (lty_listcons t')) 1).

  intros.
  
  apply check_instr_weaken with
    (interp_cont p (cm_unset e v) 1 && t_slot v0 [[ t0 |p] && t_slot v1 [[ t1 |p])
    (interp_cont p (cm_unset e v) 1 && t_slot v (t_pair [[ t0 |p] [[ t1 |p])).
  apply sub_inter_right.
  apply sub_inter_right.
  apply derives_axiom.
  unfold t_sub; repeat apply modal_N.
  simpl_derives.
  apply cm_unset_interp.
  apply derives_axiom.
  apply (cm_get_interp p e v0 t0).
  auto.
  apply derives_axiom.
  apply (cm_get_interp p e v1 t1).
  auto.
  cut_with (interp_cont p (cm_unset e v) 1 && t_slot v (t_listcons [[ t' |p]) >=>
            interp_cont p (cm_set e v (lty_listcons t')) 1).
  cut_with (interp_cont p (cm_unset e v) 1 && t_slot v (t_pair [[ t0 |p] [[ t1 |p]) >=>
            interp_cont p (cm_unset e v) 1 && t_slot v (t_listcons [[ t' |p])).
  clear_at 2.
  apply sub_trans.
  apply derives_axiom.
  apply sub_inter_right.
  apply sub_inter_left1.
  apply sub_refl.
  apply sub_inter_left2.
  apply_d slot_sub.
  unfold t_listcons.
  apply sub_inter_right.
  unfold t_sub.
  rewrite list_unfold.
  repeat apply modal_N; simpl_derives.
  apply union_intro2.
  cut_with (t_pair [[ t0 |p] [[ t1 |p] >=>
     t_pair [[ t' |p] (|>t_list [[ t' |p])).
  apply sub_project_herenow.
  clear_at 0.
  apply t_pair_sub_congr.
  auto.
  simpl in H2.
  cut_with (t_list [[ t' |p] >=> |>t_list [[ t' |p]).
  cut_with ([[ t1 |p] >=> t_list [[ t' |p]).
  apply sub_trans.
  apply derives_axiom; apply H2.
  unfold t_sub; repeat apply modal_N; simpl_derives.
  destruct (interp_ignore_nec p (1+list_ty_depth (lty_list t'))).
  destruct (H4 (lty_list t')); auto.
  apply t_pair_sub_congr.
  auto.
  simpl in H2.
  cut_with (t_list [[ t' |p] >=> |>t_list [[ t' |p]).
  cut_with ([[ t1 |p] >=> t_list [[ t' |p]).
  apply sub_trans.
  apply derives_axiom; apply H2.
  unfold t_sub; repeat apply modal_N; simpl_derives.
  destruct (interp_ignore_nec p (1+list_ty_depth (lty_list t'))).
  destruct (H4 (lty_list t')); auto.
  apply derives_axiom.
  unfold t_sub; repeat apply modal_N.
  simpl_derives.
  rewrite cm_set_unset.
  apply (cm_set_interp p (cm_unset e v) v (lty_listcons t')).  

  apply check_instr_cons.
  apply lm_pt_every.
  apply lm_ok_nec_every.
  destruct (interp_ignore_nec p (1+list_ty_depth t0)).
  destruct (H4 t0); auto.
  apply ignores_incl with (Sel_I::Sel_S::nil).
  red; simpl; intuition.
  auto.
  destruct (interp_ignore_nec p (1+list_ty_depth t1)).
  destruct (H4 t1); auto.
  apply ignores_incl with (Sel_I::Sel_S::nil).
  red; simpl; intuition.
  auto.
  destruct (interp_ignore_nec p (1+contmap_depth (cm_unset e v))).
  destruct (H3 (cm_unset e v) 1%positive); auto.
  apply ignores_incl with (Sel_I::Sel_V::nil).
  red; simpl; intuition.
  auto.
  apply cm_unset_not_in_dom.
  destruct (interp_ignore_nec p (1+contmap_depth (cm_unset e v))).
  destruct (H3 (cm_unset e v) 1%positive); auto.
  destruct (interp_ignore_nec p (1+list_ty_depth t0)).
  destruct (H4 t0); auto.
  destruct (interp_ignore_nec p (1+list_ty_depth t1)).
  destruct (H4 t1); auto.
 Qed.  

Lemma lm_check_instr_fetch_0 : forall p pt e v v0 t,
   cm_get e v = Some (lty_listcons t) ->
   check_instr p (lm_pt p pt) (lm_ok p pt) (interp_cont p e 1)
     (instr_fetch_field v 0 v0) (interp_cont p (cm_set e v0 t) 1).

  intros.
  apply check_instr_weaken with
    (interp_cont p (cm_unset e v0) 1 && t_slot v  (t_pair (|>[[ t |p]) (|>t_list [[ t |p])))
    (interp_cont p (cm_unset e v0) 1 && t_slot v0 [[ t |p]).
  apply sub_inter_right.
  apply derives_axiom.
  unfold t_sub; repeat apply modal_N.
  simpl_derives.
  apply cm_unset_interp.
  apply derives_axiom.
  cut_with (t_slot v [[ lty_listcons t |p] >=> 
            t_slot v (t_pair (|>[[ t |p]) (|>t_list [[ t |p]))).
  cut_with (interp_cont p e 1 >=> t_slot v [[ lty_listcons t |p]).
  apply sub_trans.
  clear_at 0.
  apply (cm_get_interp p e v (lty_listcons t)).
  auto.
  apply_d slot_sub.
  simpl.
  unfold t_listcons.
  apply sub_inter_left2.
  apply t_pair_sub_congr.
  unfold t_sub; repeat apply modal_N.
  simpl_derives.
  destruct (interp_ignore_nec p (1+list_ty_depth t)).  
  destruct (H1 t); auto.
  apply sub_refl.
  apply derives_axiom.
  unfold t_sub.
  repeat apply modal_N.
  simpl_derives.
  rewrite cm_set_unset.
  apply cm_set_interp.
  
  apply check_instr_fetch_0.
  apply lm_pt_every.
  destruct (interp_ignore_nec p (1+list_ty_depth t)).
  destruct (H1 t); auto.
  apply lm_ok_nec_every.
  destruct (interp_ignore_nec p (1+list_ty_depth t)).
  destruct (H1 t); auto.
  apply ignores_incl with (Sel_I::Sel_S::nil).
  red; simpl; intuition.
  auto.
  destruct (interp_ignore_nec p (1+contmap_depth (cm_unset e v0))).
  destruct (H0 (cm_unset e v0) 1%positive).
  auto.
  apply ignores_incl with (Sel_I::Sel_V::nil); auto.
  red; simpl; intuition.
  destruct (interp_ignore_nec p (1+contmap_depth (cm_unset e v0))).
  destruct (H0 (cm_unset e v0) 1%positive).
  auto.  
  auto.
  apply cm_unset_not_in_dom.
 Qed. 

Lemma lm_check_instr_fetch_1 : forall p pt e v v0 t,
   cm_get e v = Some (lty_listcons t) ->
   check_instr p (lm_pt p pt) (lm_ok p pt) (interp_cont p e 1)
     (instr_fetch_field v 1 v0) (interp_cont p (cm_set e v0 (lty_list t)) 1).

  intros.
  apply check_instr_weaken with
    (interp_cont p (cm_unset e v0) 1 && t_slot v  (t_pair (|>[[ t |p]) (|>t_list [[ t |p])))
    (interp_cont p (cm_unset e v0) 1 && t_slot v0 (t_list [[ t |p])).
  apply sub_inter_right.
  apply derives_axiom.
  unfold t_sub; repeat apply modal_N.
  simpl_derives.
  apply cm_unset_interp.
  apply derives_axiom.
  cut_with (t_slot v [[ lty_listcons t |p] >=> 
            t_slot v (t_pair (|>[[ t |p]) (|>t_list [[ t |p]))).
  cut_with (interp_cont p e 1 >=> t_slot v [[ lty_listcons t |p]).
  apply sub_trans.
  clear_at 0.
  apply (cm_get_interp p e v (lty_listcons t)).
  auto.
  apply_d slot_sub.
  simpl.
  unfold t_listcons.
  apply sub_inter_left2.
  apply t_pair_sub_congr.
  unfold t_sub; repeat apply modal_N.
  simpl_derives.
  destruct (interp_ignore_nec p (1+list_ty_depth t)).
  destruct (H1 t); auto.
  apply sub_refl.
  apply derives_axiom.
  unfold t_sub.
  repeat apply modal_N.
  simpl_derives.
  rewrite cm_set_unset.
  apply (cm_set_interp p (cm_unset e v0) v0 (lty_list t)).

  apply check_instr_fetch_1.
  apply lm_pt_every.
  apply lm_ok_nec_every.
  destruct (interp_ignore_nec p (1+list_ty_depth (lty_list t))).
  destruct (H1 (lty_list t)); auto.
  destruct (interp_ignore_nec p (1+list_ty_depth (lty_list t))).
  destruct (H1 (lty_list t)); auto.
  apply ignores_incl with (Sel_I::Sel_S::nil).
  red; simpl; intuition.
  auto.
  destruct (interp_ignore_nec p (1+contmap_depth (cm_unset e v0))).
  destruct (H0 (cm_unset e v0) 1%positive).
  auto.
  apply ignores_incl with (Sel_I::Sel_V::nil); auto.
  red; simpl; intuition.
  destruct (interp_ignore_nec p (1+contmap_depth (cm_unset e v0))).
  destruct (H0 (cm_unset e v0) 1%positive).
  auto.  
  auto.
  apply cm_unset_not_in_dom.
 Qed. 

Lemma lm_check_instr_getnil : forall p pt e v,
 check_instr p (lm_pt p pt) (lm_ok p pt) (interp_cont p e 1)
   (instr_getlabel 1%positive v) (interp_cont p (cm_set e v lty_nil) 1).

  intros.
  rewrite cm_set_unset.
  apply check_instr_weaken with
     (interp_cont p (cm_unset e v) 1)
     (interp_cont p (cm_set (cm_unset e v) v lty_nil) 1).
  apply derives_axiom.
  unfold t_sub; repeat apply modal_N.  
  simpl_derives; apply cm_unset_interp.
  apply derives_axiom; apply sub_refl.

  unfold check_instr.     
  apply forall_intro; intro i'.
  apply implies_intro.
  apply step_guard.
  simpl; intuition; subst t.
  unfold guard.
  unfold t_sub.
  rewrite nec_every_commute.
  rewrite nec_nec.
  unfold t_every; rewrite <- ignore_ignore; assumed.
  rewrite later_every_commute.
  rewrite nec_later_later.
  coalesce_modal.
  apply lm_pt_every.
  apply lm_ok_nec_every.
  apply exist_intro with (interp_cont p (cm_set (cm_unset e v) v lty_nil) 1).
  apply exist_intro with i'.
  apply (with_projecting Sel_S); intro s.
  apply exist_intro with (s#v <- (value_label 1%positive)).
  simpl_derives.
  unfold guard.
  unfold t_sub.
  rewrite later_nec_later.
  apply_d nec_later.
  assumed.
  apply steps_to_intro with s (instr_getlabel 1%positive v;;i');
    try assumed.
  apply step_getlabel.

  change
    (existT SelType Sel_S (s#v <- (value_label 1%positive)) ::
    (existT SelType Sel_I i':: nil)) with
    ((existT SelType Sel_S (s#v <- (value_label 1%positive)) :: nil) ++
     (existT SelType Sel_I i':: nil)).
  rewrite with_subst_app.
  rewrite discard_with_subst.
2:  apply with_subst_ignores.
2:  apply later_ignores.
2:  simpl; intuition; discriminate.
2:  simpl.
2:  destruct (interp_ignore_nec p (1 + (contmap_depth (cm_set (cm_unset e v) v lty_nil)))).
2:  destruct (H (cm_set (cm_unset e v) v lty_nil) 1%positive); auto.
2:  apply ignores_incl with (Sel_I::Sel_V::nil).
2:  red; simpl; intuition.
2:  apply H1.
  generalize (cm_unset_not_in_dom p e v).
  unfold not_in_dom, with_var_upd; intros.
  cut_with (t_comp_prop Sel_S (fun s => 
      with_subst (existT SelType Sel_S (s#v <- (value_label 1%positive)) :: nil)
        (interp_cont p (cm_unset e v) 1))).
  elim_at 0 (comp_prop_elim Sel_S s); [ assumed | ].
  do 5 clear_at 2.
  in_goal rewrite <- with_subst_dup.
  move_at 1.
  apply with_subst_shift.
  simpl; auto.
  rewrite <- with_subst_app.
  simpl.
  rewrite with_subst_normalize.
  simpl.
  rewrite with_subst_nil.
  cut_with (interp_cont p (cm_unset e v) 1).
  clear_at 2.
  apply with_subst_intro.
  assumed.
  simpl.
  rewrite with_subst_nil.
  cut_with (interp_cont p (cm_set (cm_unset e v) v lty_nil) 1).
  do 2 clear_at 1.
  destruct (interp_ignore_nec p (1 + (contmap_depth (cm_set (cm_unset e v) v lty_nil)))).
  destruct (H0 (cm_set (cm_unset e v) v lty_nil) 1%positive).
  auto.
  auto.
  cut_with (t_slot v [[ lty_nil |p]).
  clear_at 2; move_at 1.
  refine (cm_set_interp p (cm_unset e v) v lty_nil).
  clear_at 0.
  unfold t_slot.
  apply exist_intro with (value_label 1%positive).
  simpl_derives.
  apply comp_prop_intro with (s#v <- (value_label 1%positive)).
  assumed.
  apply const_intro.
  rewrite get_set_same.
  auto.
  clear_at 0.
  cut_with (t_top).
  apply (with_projecting Sel_V); intro a.
  apply with_subst_shift.
  simpl; auto.
  clear_at 1.
  rewrite with_subst_nil.
  unfold t_nil.
  simpl.
  unfold t_just; assumed.
  apply top_intro.
  move_at 1.
  apply with_subst_elim.
  assumed.
  simpl.
  rewrite with_subst_nil.
  assumed.
  weaken_with H.
 Qed.
  

Lemma lm_check_instr_getlabel : forall p pt l c e v,
   pt # l = Some c ->
   check_instr p (lm_pt p pt) (lm_ok p pt) (interp_cont p e 1)
     (instr_getlabel l v) (interp_cont p (cm_set e v (lty_cont c)) 1).

  intros.
  apply check_instr_weaken with
      (interp_cont p (cm_unset e v) 1)
     ((interp_cont p (cm_unset e v) 1) &&
       (t_slot v (t_cont p (interp_cont p c 1)))).
  apply derives_axiom.
  unfold t_sub; repeat apply modal_N.
  simpl_derives.
  apply cm_unset_interp.

  apply derives_axiom.
  unfold t_sub; repeat apply modal_N.
  simpl_derives.
  rewrite cm_set_unset.
  refine (cm_set_interp p (cm_unset e v) v (lty_cont c)).
  apply check_instr_getlabel.
  apply lm_pt_every.
  apply lm_ok_nec_every.
  destruct (interp_ignore_nec p (1 + (contmap_depth (cm_unset e v)))).
  destruct (H0 (cm_unset e v) 1%positive); auto.
  apply cm_unset_not_in_dom.
  destruct (interp_ignore_nec p (1 + (contmap_depth (cm_unset e v)))).
  destruct (H0 (cm_unset e v) 1%positive); auto.
  red.
  apply ignores_incl with (Sel_I::Sel_V::nil).
  red; simpl; intuition.
  auto.
  apply lm_pt_labelguard; auto.
  apply lm_ok_labelblock with c; auto.
 Qed.


Lemma lm_check_instr_branch_on_nil : forall p pt l e v,
  cm_get e v = Some lty_nil ->
  check_branch p pt e l ->
  check_instr p (lm_pt p pt) (lm_ok p pt) (interp_cont p e 1)
     (instr_branch_if_nil v l) (interp_cont p e 1).

  intros.
  red in H0.
  case_eq (pt#l); intros; rewrite H1 in H0; try contradiction.
  apply check_instr_weaken with
    ((interp_cont p e 1) && t_slot v t_nil)
    (interp_cont p e 1).
  apply sub_inter_right.
  apply derives_axiom; apply sub_refl.
  apply derives_axiom; apply (cm_get_interp p e v lty_nil).
  auto.
  apply derives_axiom; apply sub_refl.
  apply check_instr_branch_nil.  
  apply lm_pt_every.
  apply lm_ok_nec_every.
  cut_with (labelguard p l (interp_cont p c 1)).
  unfold labelguard.
  simpl_derives.
  elim_at 0 (forall_elim _ x).
  cut_with (guard p (interp_cont p c 1) x >=> guard p (interp_cont p e 1 && t_slot v t_nil) x).
  clear_at 2.
  clear_at 2.
  move_at 1.
  apply sub_trans.
  apply derives_axiom.
  unfold t_sub; repeat apply modal_N.
  apply_d guard_sub.
  apply sub_inter_left1.
  auto.
  apply lm_pt_labelguard.
  auto.
  destruct (interp_ignore_nec p (1+contmap_depth e)).
  destruct (H2 e 1%positive); auto.
  apply lm_ok_labelblock with c; auto.
  destruct (interp_ignore_nec p (1+contmap_depth e)).
  destruct (H2 e 1%positive); auto.
 Qed.  

Lemma lm_check_instr_branch_on_listcons : forall p pt l e v t,
   cm_get e v = Some (lty_listcons t) ->
   check_instr p (lm_pt p pt) (lm_ok p pt) (interp_cont p e 1)
     (instr_branch_if_nil v l) (interp_cont p e 1).

  intros.
  cut (interp_cont p e 1 = interp_cont p e 1 && t_slot v t_anypair).
  intros.
  apply check_instr_weaken with
      (interp_cont p e 1 && (t_slot v t_anypair))
      (interp_cont p e 1 && (t_slot v t_anypair)).
  unfold t_sub; rewrite <- H0.
  apply derives_axiom.
  repeat apply modal_N; simpl_derives; assumed.
  unfold t_sub; rewrite <- H0.
  apply derives_axiom.
  repeat apply modal_N; simpl_derives; assumed.
  apply check_instr_branch_pair.
  apply lm_pt_every.
  apply lm_ok_nec_every.
  destruct (interp_ignore_nec p (1+contmap_depth e)).
  destruct (H1 e 1%positive); auto.
  destruct (interp_ignore_nec p (1+contmap_depth e)).
  destruct (H1 e 1%positive); auto.
  apply equ_equal.
  unfold t_equ; simpl_derives.
  apply sub_inter_right.
  apply sub_refl.
  cut_with (t_slot v [[ lty_listcons t |p] >=> t_slot v t_anypair).
  cut_with (interp_cont p e 1 >=> t_slot v [[ lty_listcons t |p]).
  apply sub_trans.
  clear_at 0.
  apply (cm_get_interp p e v (lty_listcons t)).
  auto.
  apply_d slot_sub.  
  simpl.
  unfold t_listcons.
  apply sub_inter_left2.
  unfold t_anypair.
  unfold t_sub; repeat apply modal_N.
  simpl_derives.
  apply exist_intro with [[ t |p].
  apply exist_intro with (|> t_list [[ t |p]).
  rewrite <- interp_valuepred.
  change (t_list [[ t |p]) with ([[ lty_list t |p]).
  rewrite <- later_interp_valuepred.
  assumed.
  apply sub_inter_left1.
  apply sub_refl.
 Qed.


Lemma lm_check_instr_branch_on_list : forall p pt l e v t,
  cm_get e v = Some (lty_list t) ->
  check_branch p pt (cm_set e v lty_nil) l ->
  check_instr p (lm_pt p pt) (lm_ok p pt) (interp_cont p e 1)
     (instr_branch_if_nil v l) (interp_cont p (cm_set e v (lty_listcons t)) 1).

  intros.
  red in H0.
  case_eq (pt#l); intros; rewrite H1 in H0; try contradiction.
  apply check_instr_weaken with
      (interp_cont p e 1 && (t_slot v (t_nil || t_anypair)))
      (interp_cont p e 1 && (t_slot v t_anypair)).
  apply sub_inter_right.
  apply derives_axiom; apply sub_refl.
  apply derives_axiom.
  cut_with (t_slot v [[ lty_list t |p] >=>
            t_slot v (t_nil || t_anypair)).
  cut_with (interp_cont p e 1 >=> t_slot v [[ lty_list t |p]).
  apply sub_trans.
  clear_at 0.
  apply (cm_get_interp p e v (lty_list t)).
  auto.
  apply_d slot_sub.
  simpl.
  unfold t_sub.
  rewrite list_unfold.
  repeat apply modal_N.
  simpl_derives.
  apply union_intro1; assumed.
  apply union_intro2.
  unfold t_anypair.
  apply exist_intro with [[ t |p].
  apply exist_intro with (|> t_list [[ t |p]).
  rewrite <- (interp_valuepred p t).
  cut (|> t_list [[t |p] = t_valuepred (|>t_list [[t |p])).
  intro H2.
  rewrite <- H2.
  assumed.
  change (t_list [[ t |p]) with ([[ lty_list t |p]).
  rewrite <- later_interp_valuepred.
  auto.
  apply derives_axiom.
  unfold t_sub.
  repeat apply modal_N.
  simpl_derives.
  cut_with (t_slot v (t_listcons [[ t |p])).
  clear_at 2.
  move_at 1.
  refine (cm_set_interp p e v (lty_listcons t)).
  cut_with (t_slot v (t_list [[ t |p])).
  clear_at 1.
  unfold t_slot.
  simpl_derives.
  apply (with_projecting Sel_S); intro s.
  elim_at 1 (comp_prop_elim Sel_S s); [ assumed | ].
  elim_at 3 (comp_prop_elim Sel_S s); [ assumed | ].
  simpl_derives.
  replace x0 with x; [ | congruence ].
  apply exist_intro with x.
  simpl_derives.
  apply comp_prop_intro with s; [ assumed | ].
  apply const_intro; auto.
  coalesce_modal.
  unfold t_listcons.
  simpl_derives; try assumed.
  rewrite list_unfold.
  simpl_derives.
  cut_with t_bot.
  elim_at 0 bot_elim.
  unfold t_nil, t_anypair.
  simpl_derives.
  unfold t_just, t_pair.
  elim_at 0 (comp_prop_elim Sel_V (value_label 1%positive)).
  assumed.
  assumed.
  rewrite <- list_unfold.  
  assumed.
  cut_with (interp_cont p e 1 >=> t_slot v [[ lty_list t |p]).
  clear_at 2.
  apply sub_project_herenow.  
  apply derives_axiom.
  apply cm_get_interp.  
  auto. 

  apply check_instr_branch.
  apply lm_pt_every.
  apply lm_ok_nec_every.
  cut_with (labelguard p l (interp_cont p c 1)).
  clear_at 1.
  clear_at 1.
  unfold labelguard.
  simpl_derives.
  elim_at 0 (forall_elim _ x).
  cut_with (guard p (interp_cont p c 1) x >=> 
            guard p (interp_cont p e 1 && t_slot v t_nil) x).
  move_at 1.
  apply sub_trans.
  clear_at 0.
  unfold t_sub.
  repeat apply modal_N.
  apply_d guard_sub.
  cut_with (interp_cont p (cm_set e v lty_nil) 1 >=> interp_cont p c 1).
  cut_with (interp_cont p e 1 && t_slot v t_nil >=> interp_cont p (cm_set e v lty_nil) 1).
  apply sub_trans.
  apply derives_axiom.
  unfold t_sub.
  repeat apply modal_N.
  simpl_derives.
  apply (cm_set_interp p e v lty_nil).
  auto.
  apply lm_pt_labelguard.
  auto.
  apply lm_ok_labelblock with c; auto.
  destruct (interp_ignore_nec p (1+contmap_depth e)).
  destruct (H2 e 1%positive); auto.
  destruct (interp_ignore_nec p (1+contmap_depth e)).
  destruct (H2 e 1%positive); auto.
 Qed.
